%%  A Bimodal Lognormal Distribution Model for the Prediction of COVID-19 Deceases
%
%   by Paolo S. VALVO
%
%   University of Pisa
%   Department of Civil and Industrial Engineering
%   Largo Lucio Lazzarino
%   I-56122 Pisa (PI)
%   Italy
%
%   E-mail: p.valvo@ing.unipi.it
%   Tel.: +39-050-2218223
%   Home page: http://www2.ing.unipi.it/paolovalvo/index.html
%
%   Matlab script file name: COVID19_BiLog_World.m
%   Data source: https://ourworldindata.org/coronavirus-source-data
%   Last update: 22 November 2020
%   ----------------------------------------------------------------------

close all
clear variables


%%  Setting of general flags
%   Download data from web
%   if download_data == 0 then data will be read from local CSV file owid-covid-data.csv
%   if download_data == 1 then data will be downloaded from OWID website https://ourworldindata.org/coronavirus-source-data
download_data = 0;          % 0 = no; 1 = yes

%   Continent name in output filenames
%   if continent_file_name == 0 then PNG files will be named only with country names
%   if continent_file_name == 1 then continent names will be added to country names
continent_file_name = 1;	% 0 = no; 1 = yes

%   Plot all locations
%   if plot_all == 0 then plots will be traced only for the locations in the list
%   if plot_all == 1 then plots will be traced for all of the locations in the dataset
plot_all = 1;               % 0 = no; 1 = yes
plot_locations_list = [ "Africa", "Asia", "Europe", "North America", "Oceania", "South America", "World" ];

%   Plot unconverged countries
%   if plot_unconverged == 0 then plots will be traced according to plot_all and plot_locations_list only
%   if plot_unconverged == 1 then plots will added for the countries where the alogorithm does not converge
plot_unconverged = 1;       % 0 = no; 1 = yes

%   Plot error bars
%   if plot_error_bars == 0 then no error bars will be plotted
%   if plot_error_bars == 1 then error bars will be plotted in continents and World extrapolations
plot_error_bars = 1;        % 0 = no; 1 = yes
error_rate_lower = 0.025;	% 2.5%/15 days
error_rate_upper = 0.050;   % 5.0%/15 days

%   Plot scatter plots
%   if plot_scatter == 0 then no scatter plotswill be plotted
%   if plot_scatter == 1 then scatter plots will be plotted
%   Note: actual data should be available in CSV file between the end of fitting and the forecast dates
plot_scatter = 0;           % 0 = no; 1 = yes


%%  Definition of important dates
%   Start date
start_date =       datetime(2019,12,31);	% origin of time (day 1)

%   End date of fitting range
fit_end_date =     datetime(2020,09,30)     % later than start date

%   Forecast date
forecast_date =    datetime(2020,12,31)     % later than end of fitting date

%   Plot date range
plot_start_date =  datetime(2019,12,31);	% greater than or equal to start_date
plot_end_date =    datetime(2020,12,31);	% greater than or equal to extrapolation date


%%  Reading of data

t0 = cputime;

if download_data == 0
    fprintf("\nReading data from file...\n")
    OWID_data = readtable('owid-covid-data.csv');
elseif download_data == 1
    fprintf("\nReading data from web...\n")
    OWID_data = webread("https://covid.ourworldindata.org/data/owid-covid-data.csv");
end

t1 = cputime;
fprintf("\n- elapsed CPU time: %.3f seconds\n", t1 - t0)


%%  Preparation of data
fprintf("\nPreparing data for fitting...\n")

%	Extraction of data start and end dates
rows = (OWID_data.location == "World");
temp_dates = OWID_data.date(rows);

data_start_date = temp_dates(1);
data_end_date = temp_dates(end);
data_dates = data_start_date:data_end_date;
N_data_days = length(data_dates);
data_days = (1:N_data_days) + days(data_start_date - start_date);

plot_dates = plot_start_date:plot_end_date;
N_plot_days = length(plot_dates);
plot_days = (1:N_plot_days) + days(plot_start_date - start_date);
clear temp_dates

%	Extraction of location names
continents = string(unique(OWID_data.continent));
continents(strcmp('',continents)) = [];
N_continents = length(continents);

countries = string(unique(OWID_data.location));
countries(strcmp('',countries)) = [];
[~, i_World] = ismember("World", countries);
countries = [ countries(1:i_World-1); countries(i_World+1:end) ];
N_countries = length(countries);

locations = [ countries; continents; "World" ];
N_locations = length(locations);

%	Definition of countries to be fitted
fit_countries = 1:N_countries;

%	Initialisation of variables
continent = zeros(1,N_countries);

population            = zeros(1,N_locations);
outbreak_start_day    = zeros(1,N_locations);
fit_end_day           = zeros(1,N_locations);
daily_deaths_peak_day = zeros(N_locations,2);
daily_deaths_peak     = zeros(N_locations,2);

total_deaths       = zeros(N_locations,N_data_days);
daily_deaths       = zeros(N_locations,N_data_days);

total_deaths_model = zeros(N_locations,N_plot_days);
daily_deaths_model = zeros(N_locations,N_plot_days);
forecast_deaths    = zeros(N_locations,2);

A        = zeros(1,N_countries);
M1       = zeros(1,N_countries);
sigma_t1 = zeros(1,N_countries);
M2       = zeros(1,N_countries);
sigma_t2 = zeros(1,N_countries);
alpha    = zeros(1,N_countries);
resnorm  = zeros(1,N_countries);
exitflag = zeros(1,N_countries);

%	Extraction of data for fitting
for i = 1:N_countries
    rows = (OWID_data.location == string(locations(i)));
    temp_continent    = OWID_data.continent(rows);
    temp_population   = OWID_data.population(rows);
    temp_dates        = OWID_data.date(rows);
    temp_total_deaths = OWID_data.total_deaths(rows);
    temp_daily_deaths = OWID_data.new_deaths(rows);

%   Extraction of continent number
    [~, continent(i)] = ismember(string(temp_continent(1)), continents);

%   Extraction of population data
    if isnan(temp_population(1))
        population(i) = 0;
    else
        population(i) = temp_population(1);
    end

%   Extraction of deaths data
    for j = 1:length(temp_dates)
        k = days(temp_dates(j) - start_date) + 1;
        if isnan(temp_total_deaths(j))
            if k > 1
                total_deaths(i,k) = total_deaths(i,k-1);
            else
                total_deaths(i,k) = 0;
            end
        else
            total_deaths(i,k) = temp_total_deaths(j);
        end
        if isnan(temp_daily_deaths(j))
            daily_deaths(i,k) = 0;
        else
            daily_deaths(i,k) = temp_daily_deaths(j);
        end
    end

%   Outbreak start day is the day the first deaths are reported
    [~, outbreak_start_day(i)] = max(total_deaths(i,:) > 0);

%   Fit end day cannot be later than last data day in a country
    fit_end_day(i) = min(days(temp_dates(end) - start_date), days(fit_end_date - start_date)) + 1;

%   Daily deaths peaks = first and second maxima in country data series
    [daily_deaths_peak(i,1), daily_deaths_peak_day(i,1)] = max(daily_deaths(i,1:fit_end_day(i)));
    [daily_deaths_peak(i,2), daily_deaths_peak_day(i,2)] = max([daily_deaths(i,1:(daily_deaths_peak_day(i,1) - 1)) 0 daily_deaths(i,(daily_deaths_peak_day(i,1) + 1):fit_end_day(i))]);
    if daily_deaths_peak(i,2) == 0
        daily_deaths_peak(i,2) = daily_deaths_peak(i,1);
        daily_deaths_peak_day(i,2) = daily_deaths_peak_day(i,1);
    end

    clear temp_continent temp_population temp_dates temp_total_deaths temp_daily_deaths
end

t2 = cputime;
fprintf("\n- elapsed CPU time: %.3f seconds\n", t2 - t1)


%%  Fitting of data
fprintf("\nFitting data...\n")

%   Fitting options
options = optimoptions('lsqcurvefit','FunctionTolerance',1e-9,'OptimalityTolerance',1e-9,'StepTolerance',1e-9,'MaxFunctionEvaluations',100000,'MaxIterations',10000);

%   Fitting of data
forecast_day = days(forecast_date - start_date) + 1;
t_max = 367;
unconverged_countries = double.empty;

for i = fit_countries
    fprintf("\n- fitting data for country No. %u: %s\n", i, locations(i))

    fit_days = 1:fit_end_day(i);
	if (total_deaths(i,fit_end_day(i)) > 0)

        A_min = 1.1*total_deaths(i,fit_end_day(i));
        A_gss = max( 6*total_deaths(i,daily_deaths_peak_day(i,1)), 2*total_deaths(i,fit_end_day(i)));
        A_max = max(12*total_deaths(i,daily_deaths_peak_day(i,1)), 4*total_deaths(i,fit_end_day(i)));

        M1_min = 1;
        M1_gss = daily_deaths_peak_day(i,1);
        M1_max = t_max;

        sigma_t1_min = 0;
        sigma_t1_gss = 1;
        sigma_t1_max = t_max;

        M2_min = 1;
        M2_gss = daily_deaths_peak_day(i,2);
        M2_max = t_max;

        sigma_t2_min = 0;
        sigma_t2_gss = 1;
        sigma_t2_max = t_max;

        alpha_min = 0.00;
        alpha_gss = 1/(1 + daily_deaths_peak(i,2)/daily_deaths_peak(i,1));
        alpha_max = 1.00;

        param_min = [ A_min M1_min sigma_t1_min M2_min sigma_t2_min alpha_min ];
        param_gss = [ A_gss M1_gss sigma_t1_gss M2_gss sigma_t2_gss alpha_gss ];
        param_max = [ A_max M1_max sigma_t1_max M2_max sigma_t2_max alpha_max ];

        [param,resnorm(i),residual,exitflag(i),output] = lsqcurvefit(@bilognormal,param_gss,fit_days,total_deaths(i,fit_days),param_min,param_max,options);
        if exitflag(i) < 1
            unconverged_countries = [ unconverged_countries i ];
        end
        if param(2) > param(4)
            [param(2), param(4)] = deal(param(4),param(2));
            [param(3), param(5)] = deal(param(5),param(3));
            param(6) = 1 - param(6);
        end

        A(i)        = param(1);
        M1(i)       = param(2);
        sigma_t1(i) = param(3);
        M2(i)       = param(4);
        sigma_t2(i) = param(5);
        alpha(i)    = param(6);

        total_deaths_model(i,:) = bilognormal(param, plot_days);
    else
        fprintf("\nlsqcurvefit not used: 0 deaths reported, assuming null functions.\n")

        A(i)        = 0;
        M1(i)       = 0;
        sigma_t1(i) = 0;
        M2(i)       = 0;
        sigma_t2(i) = 0;
        alpha(i)    = 0;
        resnorm(i)  = 0;
        exitflag(i) = 1;

        total_deaths_model(i,:) = 0;
	end
    daily_deaths_model(i,:) = diff([0 total_deaths_model(i,:)]);
    forecast_deaths(i,1) = round(total_deaths_model(i,forecast_day));
    forecast_deaths(i,2) = round(A(i));
end

%   Printing of the list of unconverged countries
fprintf("\n- unconverged countries:\n\n")
if isempty(unconverged_countries)
    fprintf("None\n")
else
    fprintf("%s\n", countries(unconverged_countries))
end

t3 = cputime;
fprintf("\n- elapsed CPU time: %.3f seconds\n", t3 - t2)


%%  Computation of continental and World summaries
fprintf("\nComputing continental and World summaries...\n")

%   Computation of continent data
fprintf("\n- computing continent data...\n")

for j = 1:N_continents
    i = N_countries + j;
    outbreak_start_day(i) = min(outbreak_start_day(continent == j));
	fit_end_day(i) = max(fit_end_day(continent == j));

    population(i) = sum(population(continent == j));
    total_deaths(i,:) = sum(total_deaths(continent == j,data_days));
    daily_deaths(i,:) = sum(daily_deaths(continent == j,data_days));

    [daily_deaths_peak(i,1), daily_deaths_peak_day(i,1)] = max(daily_deaths(i,1:fit_end_day(i)));
    [daily_deaths_peak(i,2), daily_deaths_peak_day(i,2)] = max([daily_deaths(i,1:(daily_deaths_peak_day(i,1) - 1)) 0 daily_deaths(i,(daily_deaths_peak_day(i,1) + 1):fit_end_day(i))]);
    if daily_deaths_peak(i,2) == 0
        daily_deaths_peak(i,2) = daily_deaths_peak(i,1);
        daily_deaths_peak_day(i,2) = daily_deaths_peak_day(i,1);
    end

    total_deaths_model(i,:) = sum(round(total_deaths_model(continent == j,plot_days)));
    daily_deaths_model(i,:) = sum(daily_deaths_model(continent == j,plot_days));
	forecast_deaths(i,:) = sum(forecast_deaths(continent == j,:));
end

%   Computation of World data
fprintf("\n- computing World data...\n")

outbreak_start_day(end) = min(outbreak_start_day(1:N_countries));
fit_end_day(end) = max(fit_end_day(1:N_countries));

population(end) = sum(population(1:N_countries));
total_deaths(end,:) = sum(total_deaths(1:N_countries,data_days));
daily_deaths(end,:) = sum(daily_deaths(1:N_countries,data_days));

[daily_deaths_peak(end,1), daily_deaths_peak_day(end,1)] = max(daily_deaths(end,1:fit_end_day(end)));
[daily_deaths_peak(end,2), daily_deaths_peak_day(end,2)] = max([daily_deaths(end,1:(daily_deaths_peak_day(end,1) - 1)) 0 daily_deaths(end,(daily_deaths_peak_day(end,1) + 1):fit_end_day(end))]);
if daily_deaths_peak(end,2) == 0
	daily_deaths_peak(end,2) = daily_deaths_peak(end,1);
    daily_deaths_peak_day(end,2) = daily_deaths_peak_day(end,1);
end

total_deaths_model(end,:) = sum(round(total_deaths_model(fit_countries,plot_days)));
daily_deaths_model(end,:) = sum(daily_deaths_model(fit_countries,plot_days));
forecast_deaths(end,:) = sum(forecast_deaths(fit_countries,:));

t4 = cputime;
fprintf("\n- elapsed CPU time: %.3f seconds\n", t4 - t3)


%%  Saving of results
fprintf("\nSaving results to CSV file...\n")

results_folder = strcat("BiLog-", string(datetime(fit_end_date,'Format','dd MMM')), "-", string(datetime(forecast_date,'Format','dd MMM yyyy')));
if not(isfolder(results_folder))
    mkdir(results_folder)
end

header = "continent,country,population,outbreak_start_date,fit_end_date,actual_deaths,forecast_date,forecast_deaths,A,M1,sigma_t1,M2,sigma_t2,alpha,resnorm,exitflag";
file_id = fopen(strcat(results_folder, "\COVID19_BiLog_World.csv"),'w');
fprintf(file_id,'%s\n',header);
continent_names = [ "None"; continents ];
for i = 1:N_countries
    fprintf(file_id,'%s,%s,%f,%s,%s,%f,%s,%f,%f,%f,%f,%f,%f,%f,%f,%f\n',[continent_names(continent(i) + 1),countries(i),population(i),string(datetime(start_date + outbreak_start_day(i) - 1,'Format','yyyy-MM-dd')),string(datetime(start_date + fit_end_day(i) - 1,'Format','yyyy-MM-dd')),total_deaths(i,fit_end_day(i)),string(datetime(start_date + forecast_day - 1,'Format','yyyy-MM-dd')),forecast_deaths(i,1),A(i),M1(i),sigma_t1(i),M2(i),sigma_t2(i),alpha(i),resnorm(i),exitflag(i)]);
end
fclose(file_id);

t5 = cputime;
fprintf("\n- elapsed CPU time: %.3f seconds\n", t5 - t4)


%%  Plotting of results
fprintf("\nPlotting...\n")

if plot_all == 0
    plot_locations = double.empty;
    for i = 1:N_locations
        if ismember(locations(i), plot_locations_list)
            plot_locations = [ plot_locations i ];
        end
    end
	if (plot_unconverged == 1) && not(isempty(unconverged_countries))
        for i = unconverged_countries
            if not(ismember(i, plot_locations))
                plot_locations = [ plot_locations i ];
            end
        end
    end
elseif plot_all == 1
	plot_locations = 1:N_locations;
end

for i = plot_locations
    last_data_date_str = string(datetime(start_date + fit_end_day(i) - 1,'Format','dd MMM yyyy'));
    peak_date_str = string(datetime(start_date + daily_deaths_peak_day(i,1) - 1,'Format','dd MMM yyyy'));
    forecast_date_str = string(datetime(forecast_date,'Format','dd MMM yyyy'));

    if (continent_file_name == 1) && (i <= N_countries)
        file_name_str = strcat(results_folder, "\", continent_names(continent(i) + 1), "-", locations(i));
    else
        file_name_str = strcat(results_folder, "\", locations(i));
    end
    
    figure(3*i-2)
    plot(plot_dates(1:1:fit_end_day(i)),daily_deaths(i,1:1:fit_end_day(i)),'bo-')
    xlim([plot_dates(1) plot_dates(end)])
	xticks([datetime(2020,1:2:12,1)])
    title(strcat("Daily COVID-19 deaths: ", locations(i)))
    hold on
	plot(plot_dates,daily_deaths_model(i,:),'r-')
    ylim([0 1.25*max([daily_deaths_peak(i,:), max(daily_deaths_model(i,:)), 1])])
	legend(cellstr(["Data", "Model"]),'Location','northwest')
    if daily_deaths_peak_day(i,1) < N_plot_days/2
        text(daily_deaths_peak_day(i,1)+5,1.01*daily_deaths(i,daily_deaths_peak_day(i,1)),strcat("\leftarrow Daily deaths on ", peak_date_str, ": ", string(daily_deaths_peak(i,1))))
    else
        text(daily_deaths_peak_day(i,1)-5,1.01*daily_deaths(i,daily_deaths_peak_day(i,1)),strcat("Daily deaths on ", peak_date_str, ": ", string(daily_deaths_peak(i,1)),"\rightarrow"), 'HorizontalAlignment', 'right')
    end
    set(gca,'FontSize',12)
    saveas(gcf,strcat(file_name_str, "_daily.png"))
	close(3*i-2)

    figure(3*i-1)
    plot_day_shift = datenum(start_date) - 1;
    plot((1:fit_end_day(i)) + plot_day_shift,total_deaths(i,1:fit_end_day(i)),'bo')
    xlim([1 N_plot_days] + plot_day_shift)
	xticks(datenum([datetime(2020,1:2:12,1)]))
    datetick('x', 'mmm yyyy','keeplimits','keepticks')
    title(strcat("Total COVID-19 deaths: ", locations(i)))
    hold on
 	plot((1:N_plot_days) + plot_day_shift,total_deaths_model(i,:),'r-')
    text(fit_end_day(i) - 10 + plot_day_shift,1.01*total_deaths(i,fit_end_day(i)),strcat("Total deaths on ", last_data_date_str, ": ", string(total_deaths(i,fit_end_day(i))), "\rightarrow"), 'HorizontalAlignment', 'right')

    if (plot_error_bars == 1) && (i > N_countries)
        ylim([0 1.05*total_deaths_model(i,N_plot_days)*(1 + error_rate_upper*(forecast_day - fit_end_day(i))/15)])
        error_bar_days = fit_end_day(i):15:N_plot_days;
        err_lower = error_rate_lower/15*(error_bar_days - fit_end_day(i)).*total_deaths_model(i,error_bar_days);
        err_upper = error_rate_upper/15*(error_bar_days - fit_end_day(i)).*total_deaths_model(i,error_bar_days);
        errorbar(error_bar_days + plot_day_shift,total_deaths_model(i,error_bar_days),err_lower,err_upper,'k','LineStyle','none')
        legend(cellstr(["Data", "Model", "Error"]),'Location','northwest')
        text(forecast_day - 75 + plot_day_shift,1.07*forecast_deaths(i,1),strcat("Forecast deaths on ", forecast_date_str, ":"), 'HorizontalAlignment', 'right')
        text(forecast_day - 60 + plot_day_shift,1.00*forecast_deaths(i,1),strcat(string(forecast_deaths(i,1)), " + (-", ...
            string(round(forecast_deaths(i,1)*error_rate_lower*(forecast_day - fit_end_day(i))/15)), ", +", ...
            string(round(forecast_deaths(i,1)*error_rate_upper*(forecast_day - fit_end_day(i))/15)), ") \rightarrow"), 'HorizontalAlignment', 'right')
    else
        ylim([0 1.25*max(total_deaths_model(i,N_plot_days), 1)])
        legend(cellstr(["Data", "Model"]),'Location','northwest')
        text(forecast_day - 15 + plot_day_shift,1.00*forecast_deaths(i,1),strcat("Forecast deaths on ", forecast_date_str, ": ", string(forecast_deaths(i,1)), "\rightarrow"), 'HorizontalAlignment', 'right')
    end
    set(gca,'FontSize',12)
    saveas(gcf,strcat(file_name_str, "_total.png"))
	close(3*i-1)

    if plot_scatter == 1
        figure(3*i)
        scatter_days = fit_end_day(i):forecast_day;
%        scatter_days = fit_end_day(i) + [ 0 15 31 46 62 77 92]
        scatter(total_deaths_model(i,scatter_days),total_deaths(i,scatter_days),'bo')
        h1 = lsline;
        h1.LineWidth = 2;
        h1.Color = 'r';
        axis equal
        title(strcat("Total COVID-19 deaths: ", locations(i)))
        xlabel("Predicted deaths (No.)")
        ylabel("Actual deaths (No.)")
        y_slope = (h1.YData(2) - h1.YData(1))/(h1.XData(2) - h1.XData(1));
        y_intercept = h1.YData(1) - y_slope*h1.XData(1);
        y_mean = mean(total_deaths(i,scatter_days));
        SSres = sum((total_deaths(i,scatter_days) - (y_slope*total_deaths_model(i,scatter_days) + y_intercept)).^2);
        SStot = sum((total_deaths(i,scatter_days) - y_mean).^2);
        R2 = 1 - SSres/SStot;
        if y_intercept >= 0
            text(mean(h1.XData),0.90*mean(h1.YData),strcat("y = ", num2str(y_slope,'%.4f'), "x + ", num2str(round(y_intercept),'%i')))
        else
            text(mean(h1.XData),0.90*mean(h1.YData),strcat("y = ", num2str(y_slope,'%.4f'), "x - ", num2str(round(-y_intercept),'%i')))
        end
        text(mean(h1.XData),0.85*mean(h1.YData),strcat("R^{2} = ", num2str(R2,'%.4f')))
        grid on
        pbaspect([1 1 1])
        set(gca,'FontSize',12)
        saveas(gcf,strcat(file_name_str, "_scatter.png"))
        close(3*i)
    end
end

t6 = cputime;
fprintf("\n- elapsed CPU time: %.3f seconds\n", t6 - t5)


%%  Saving of variables
fprintf("\nSaving workspace...\n")

save(strcat(results_folder, "\Workspace.mat"))

t7 = cputime;
fprintf("\n- elapsed CPU time: %.3f seconds\n", t7 - t6)


%%  End of operations
fprintf("\nEnd of operations\n")

t8 = cputime;
fprintf("\n- total elapsed CPU time: %.3f seconds\n", t8 - t0)


%%  Bimodal lognormal distribution function
function F = bilognormal(param, t)
%
%   F(x) = A*(alpha*logncdf(t, mu1, sigma1) + (1 - alpha)*logncdf(t, mu2, sigma2)) =
%        = A*(alpha/2*(1 + erf((log(t) - mu1)/(sqrt(2)*sigma1))) +
%     + (1 - alpha)/2*(1 + erf((log(t) - mu2)/(sqrt(2)*sigma2))))
%
    A        = param(1);
    M1       = param(2);
    sigma_t1 = param(3);
    M2       = param(4);
    sigma_t2 = param(5);
    alpha    = param(6);

    r1 = roots([1 -1 0 0 -(sigma_t1/M1)^2]);
    r1 = r1(imag(r1) == 0);
    r1 = r1(r1 > 0);
    sigma1 = sqrt(log(r1));

    r2 = roots([1 -1 0 0 -(sigma_t2/M2)^2]);
    r2 = r2(imag(r2) == 0);
    r2 = r2(r2 > 0);
    sigma2 = sqrt(log(r2));

    mu1 = sigma1^2 + log(M1);
    mu2 = sigma2^2 + log(M2);

    n = length(t);
    F = zeros(1,n);
    for i = 1:n
%         F(i) = A*(alpha*logncdf(t(i), mu1, sigma1) + (1 - alpha)*logncdf(t(i), mu2, sigma2));
%   Note: the expression with erf() instead of logncdf() is about four times faster!
        F(i) = A/2*(alpha*(1 + erf((log(t(i)) - mu1)/(sqrt(2)*sigma1))) + (1 - alpha)*(1 + erf((log(t(i)) - mu2)/(sqrt(2)*sigma2))));
    end
end
